import os
import re
import functools
import operator
import numpy as np
from shapely.geometry import Point, LineString, Polygon, MultiPolygon
import pandas as pd
import geopandas as gpd

DICT_CIS_OSM = {
    "roads": {
        "osm_keys": ["highway", "name", "maxspeed", "lanes", "surface"],
        "osm_query": {
            "highway": [
                "motorway",
                "motorway_link",
                "trunk",
                "trunk_link",
                "primary",
                "primary_link",
                "secondary",
                "secondary_link",
                "tertiary",
                "tertiary_link",
                "residential",
                "road",
                "unclassified",
                "track",
            ]
        },
    },
    "main_roads": {
        "osm_keys": ["highway", "name", "maxspeed", "lanes", "surface"],
        "osm_query": {
            "highway": [
                "primary",
                "primary_link",
                "secondary",
                "secondary_link",
                "tertiary",
                "tertiary_link",
                "trunk",
                "trunk_link",
                "motorway",
                "motorway_link",
            ]
        },
    },
    "rail": {
        "osm_keys": ["railway", "name", "gauge", "electrified", "voltage"],
        "osm_query": {"railway": ["rail", "narrow_gauge"]},
    },
    "air": {
        "osm_keys": ["aeroway", "name"],
        "osm_query": {"aeroway": ["aerodrome", "terminal", "runway"]},
    },
    "telecom": {
        "osm_keys": ["man_made", "tower_type", "name"],
        "osm_query": {
            "man_made": ["mast", "communications_tower"],
            "tower_type": ["communication"],
        },
    },
    "water_supply": {
        "osm_keys": ["man_made", "name"],
        "osm_query": {
            "man_made": [
                "water_works",
                "water_well",
                "water_tower",
                "reservoir_covered",
                "storage_tank",
            ]
        },
    },
    "waste_solid": {
        "osm_keys": ["amenity", "name"],
        "osm_query": {"amenity": ["waste_transfer_station"]},
    },
    "waste_water": {
        "osm_keys": ["man_made", "name"],
        "osm_query": {"man_made": ["wastewater_plant"]},
    },
    "education": {
        "osm_keys": ["amenity", "building", "name"],
        "osm_query": {
            "building": ["school", "kindergarten", "college", "university", "library"],
            "amenity": ["school", "kindergarten", "college", "university", "library"],
        },
    },
    "healthcare": {
        "osm_keys": ["amenity", "building", "healthcare", "name"],
        "osm_query": {
            "amenity": ["hospital", "clinic", "doctors", "dentist", "pharmacy"],
            "building": ["hospital", "clinic"],
            "healthcare": [
                "pharmacy",
                "dentist",
                "physiotherapist",
                "alternative",
                "laboratory",
                "optometrist",
                "rehabilitation",
                "blood_donation",
                "birthing_center",
            ],
        },
    },
    "power": {
        "osm_keys": ["power", "voltage", "utility", "name"],
        "osm_query": {
            "power": [
                "line",
                "cable",
                "minor_line",
                "plant",
                "generator",
                "substation",
                "transformer",
                "pole",
                "portal",
                "tower",
                "terminal",
                "switch",
                "catenary_mast",
            ]
        },
    },
    "gas": {
        "osm_keys": ["man_made", "pipeline", "utility", "name", "substance", "content"],
        "osm_query": {
            "man_made": ["pipeline", "storage_tank"],
            "pipeline": ["substation"],
            "utility": ["gas"],
            "substance": ["gas"],
            "content": ["gas"],
        },
    },
    "food": {
        "osm_keys": ["amenity", "building", "name"],
        "osm_query": {
            "amenity": ["restaurant", "fast_food", "cafe", "pub", "bar"],
            "building": ["restaurant", "fast_food", "cafe", "pub", "bar"],
        },
    },
    "oil": {
        "osm_keys": ["pipeline", "man_made", "amenity", "name", "substance"],
        "osm_query": {
            "pipeline": ["substation"],
            "man_made": ["pipeline", "petroleum_well", "oil_refinery"],
            "amenity": ["fuel"],
            "substance": ["oil"],
        },
    },
    "wastewater": {
        "osm_keys": ["man_made", "amenity", "name"],
        "osm_query": {
            "man_made": ["wastewater_plant"],
            "amenity": ["waste_transfer_station"],
        },
    },
    "buildings": {
        "osm_keys": ["building", "amenity", "name"],
        "osm_query": {
            "building": [
                "yes",
                "house",
                "residential",
                "detached",
                "hut",
                "industrial",
                "shed",
                "apartments",
            ]
        },
    },
}


def _combine_columns(a, b):
    """
    Combine values from two input arguments 'a' and 'b' into a single string.
    Arguments:
    - a (str or None): Value from column 'A'.
    - b (str or None): Value from column 'B'.
    Returns:
    - str or None: A string of 'a', 'b' or combination. If both 'a' and 'b' are None, return None.
    """

    if pd.notna(a) and pd.notna(b) == False:  # if only a contains a string
        return f"{a}"
    elif pd.notna(b) and pd.notna(a) == False:  # if only b contains a string
        return f"{b}"
    elif pd.notna(a) and pd.notna(b):  # if both values contain a string
        if a == b:
            return f"{a}"
        elif a == "yes" or b == "yes":
            if a == "yes":
                return f"{b}"
            elif b == "yes":
                return f"{a}"
        else:
            return f"{a}"  # f"{a}_{b}" # assuming that value from column A contains the more detailed information
    else:
        None  # Decision point: If nones are existent, decide on what to do with Nones. Are we sure that these are education facilities? Delete them? Provide another tag to them?


def _filter_dataframe(features, column_names_lst):
    """
    Filter a GeoDataFrame by combining information from two specified columns and removing selected columns.
    Args:
        assets (geopandas.GeoDataFrame): The input GeoDataFrame containing spatial geometries and columns to filter.
        column_names_lst (list): A list of two column names whose information needs to be combined to create a new 'asset' column.
    Returns:
        geopandas.GeoDataFrame: A filtered GeoDataFrame with a new 'asset' column and selected columns dropped, and points converted to polygons.
    """
    if len(column_names_lst) == 2:
        features["object_type"] = features.apply(
            lambda row: _combine_columns(
                row[column_names_lst[0]], row[column_names_lst[1]]
            ),
            axis=1,
        )  # create new column based on tag information provided in two columns
    elif len(column_names_lst) == 3:
        features["object_type_temp"] = features.apply(
            lambda row: _combine_columns(
                row[column_names_lst[0]], row[column_names_lst[1]]
            ),
            axis=1,
        )  # create temp column based on tag information provided in two columns
        features["object_type"] = features.apply(
            lambda row: _combine_columns(
                row["object_type_temp"], row[column_names_lst[2]]
            ),
            axis=1,
        )  # create new column based on tag information provided in two columns
        column_names_lst.append("object_type_temp")
    else:
        print("Warning: column_names_lst should contain 2 or 3 items")
    features = features.drop(columns=column_names_lst, axis=1)  # drop columns
    features = _remove_contained_assets(features)

    return features


def _remove_contained_assets(features):
    """
    Process the geometry of assets, removing contained points and polygons, and converting points to polygons.
    Args:
        assets (geopandas.GeoDataFrame): Input GeoDataFrame containing asset geometries.
    Returns:
        geopandas.GeoDataFrame: Processed GeoDataFrame with updated asset geometries.
    """

    features = _remove_contained_polys(
        _remove_contained_points(features)
    )  # remove points and polygons within a (larger) polygon

    return features


def _remove_contained_points(gdf_p_mp):
    """
    from a GeoDataFrame containing points and (multi-)polygons, remove those
    points that are contained in a multipolygons entry.
    Resets the index of the dataframe.

    Parameters
    ----------
    gdf_p_mp : gpd.GeoDataFrame
        GeoDataFrame containing entries with point and (multi-)polygon geometry
    """

    gdf_p_mp = gdf_p_mp.reset_index(drop=True)

    ind_dupl = np.unique(
        gpd.sjoin(
            gdf_p_mp[gdf_p_mp.geometry.type == "Point"],
            gdf_p_mp[
                (gdf_p_mp.geometry.type == "MultiPolygon")
                | (gdf_p_mp.geometry.type == "Polygon")
            ],
            predicate="within",
        ).index
    )

    return gdf_p_mp.drop(index=ind_dupl).reset_index(drop=True)


def _remove_contained_polys(gdf):
    """
    from a GeoDataFrame containing (multi-)polygons (and potentially other
    geometries), remove those polygon entries that are already fully
    contained in another polygon entries. Removes smaller polygons within
    polygons and full duplicates, but leaves contained points untouched
    (see remove_contained_points() for this).

    Resets the index of the dataframe.

    Parameters
    ----------
    gdf : gpd.GeoDataFrame
        GeoDataFrame containing entries with (multi-)polygon geometry
    """

    gdf = gdf.reset_index(drop=True)

    contained = gpd.sjoin(
        gdf[(gdf.geometry.type == "MultiPolygon") | (gdf.geometry.type == "Polygon")],
        gdf[(gdf.geometry.type == "MultiPolygon") | (gdf.geometry.type == "Polygon")],
        predicate="contains",
    )

    subset = contained[contained.index != contained.index_right]
    to_drop = set(subset.index_right) - set(subset.index)

    return gdf.drop(index=to_drop).reset_index(drop=True)


def create_point_from_polygon(gdf):
    """
    Transforms polygons into points
    Arguments:
        gdf: A geodataframe containing a column geometry
    Returns:
    - geopandas.GeoDataFrame: The updated GeoDataFrame without polygons but with only point geometries
    """
    gdf["geometry"] = gdf["geometry"].apply(
        lambda geom: MultiPolygon([geom]) if geom.geom_type == "Polygon" else geom
    )  # convert to multipolygons in case polygons are in the df
    # gdf.loc[gdf.geom_type == 'MultiPolygon','geometry'] = gdf.loc[assets.geom_type == 'MultiPolygon'].centroid #convert polygon to point
    gdf.loc[gdf.geom_type == "MultiPolygon", "geometry"] = gdf.loc[
        gdf.geom_type == "MultiPolygon"
    ].centroid  # convert polygon to point
    return gdf


def _extract_value(text, key):
    pattern = rf'"{key}"=>"([^"]+)"'
    try:
        match = re.search(pattern, text)
        if match:
            return match.group(1)
        return None
    except:
        return None


def extract(osm_path, geom_type, osm_keys, osm_query):
    """
    Extracts data from the given osm.pbf file based on the given osm_keys and
    osm_query. The function is used by extract_osm_data() to extract critical
    infrastructure data from the osm.pbf file.
    Parameters
    ----------
    osm_path : str or Path
        location of osm.pbf file from which to parse
    geom_type : str
        one of 'points', 'lines', 'multipolygons'
    osm_keys : list
        list of osm keys to query for
    osm_query : str
        query to be executed on the osm.pbf file
    Returns
    -------
    gdf : geopandas.GeoDataFrame
        geodataframe containing the extracted data
    """
    features = gpd.read_file(osm_path, layer=geom_type, engine="pyogrio")

    for key in osm_keys:
        if key not in features.columns:
            features[key] = features["other_tags"].apply(
                lambda x: _extract_value(x, key)
            )

    # build query
    collect_indices = []
    for query_key in osm_query.keys():
        collect_indices.append(
            features[features[query_key].isin(osm_query[query_key])].index.values
        )

    # get complete list
    collect_indices = functools.reduce(operator.iconcat, collect_indices, [])

    # remove duplicates from list
    collect_indices = list(set(collect_indices))
    features = features.iloc[collect_indices]

    features = features[["osm_id", "geometry"] + osm_keys]

    # make sure we keep collect the necessary feature charactericts in same column
    # and cover the points to polygons for better damage assessment

    if len(osm_keys) < 3:
        features = _filter_dataframe(features, osm_keys)
    else:
        features = _filter_dataframe(features, osm_keys[:3])

    features.rename(columns={osm_keys[0]: "object_type"}, inplace=True)

    return features


def read_osm_data(osm_path, asset_type):
    """
    A wrapper around extract() to conveniently extract map info for a
    selection of  critical infrastructure types from the given osm.pbf file.
    No need to search for osm key/value tags and relevant geometry types.
    Parameters
    ----------
    osm_path : str or Path
        location of osm.pbf file from which to parse
    asset_type : str
        one of DICT_CIS_OSM.keys(), i.e. 'education', 'healthcare',
        'water', 'telecom', 'road', 'rail', 'air', 'gas', 'oil', 'power',
        'wastewater', 'food'
    See also
    -------
    DICT_CIS_OSM for the keys and key/value tags queried for the respective
    CIs. Modify if desired.
    """

    # features consisting in points and multipolygon results:
    if asset_type in ["healthcare", "education", "food", "buildings"]:
        gdf = pd.concat(
            [
                extract(
                    osm_path,
                    "points",
                    DICT_CIS_OSM[asset_type]["osm_keys"],
                    DICT_CIS_OSM[asset_type]["osm_query"],
                ),
                extract(
                    osm_path,
                    "multipolygons",
                    DICT_CIS_OSM[asset_type]["osm_keys"],
                    DICT_CIS_OSM[asset_type]["osm_query"],
                ),
            ]
        )

    # features consisting in points, multipolygons and lines:
    elif asset_type in ["gas", "oil", "water", "power"]:
        gdf = pd.concat(
            [
                extract(
                    osm_path,
                    "points",
                    DICT_CIS_OSM[asset_type]["osm_keys"],
                    DICT_CIS_OSM[asset_type]["osm_query"],
                ),
                extract(
                    osm_path,
                    "multipolygons",
                    DICT_CIS_OSM[asset_type]["osm_keys"],
                    DICT_CIS_OSM[asset_type]["osm_query"],
                ),
                extract(
                    osm_path,
                    "lines",
                    DICT_CIS_OSM[asset_type]["osm_keys"],
                    DICT_CIS_OSM[asset_type]["osm_query"],
                ),
            ]
        )

    # features consisting in multipolygons and lines:
    elif asset_type in ["air"]:
        gdf = pd.concat(
            [
                extract(
                    osm_path,
                    "multipolygons",
                    DICT_CIS_OSM[asset_type]["osm_keys"],
                    DICT_CIS_OSM[asset_type]["osm_query"],
                ),
                extract(
                    osm_path,
                    "lines",
                    DICT_CIS_OSM[asset_type]["osm_keys"],
                    DICT_CIS_OSM[asset_type]["osm_query"],
                ),
            ]
        )

    # features consisting in multiple datattypes, but only lines needed:
    elif asset_type in ["rail", "roads", "main_roads"]:
        gdf = pd.concat(
            [
                extract(
                    osm_path,
                    "lines",
                    DICT_CIS_OSM[asset_type]["osm_keys"],
                    DICT_CIS_OSM[asset_type]["osm_query"],
                )
            ]
        )

    # features consisting in all data types, but only points and multipolygon needed:
    elif asset_type in [
        "telecom",
        "wastewater",
        "waste_solid",
        "waste_water",
        "water_supply",
    ]:
        gdf = pd.concat(
            [
                extract(
                    osm_path,
                    "points",
                    DICT_CIS_OSM[asset_type]["osm_keys"],
                    DICT_CIS_OSM[asset_type]["osm_query"],
                ),
                extract(
                    osm_path,
                    "multipolygons",
                    DICT_CIS_OSM[asset_type]["osm_keys"],
                    DICT_CIS_OSM[asset_type]["osm_query"],
                ),
            ]
        )

    else:
        return ImportWarning("feature not in DICT_CIS_OSM. Returning empty gdf")
    return gdf



"""DamageScanner - a directe damage assessment toolkit
"""

def _check_output_path(given_args):
    """Ensures given output path exists.

    Arguments:
        *given_args* : dict, of keyword arguments.

    Returns:
        *str* : output_path, which may be empty string ('')
    """
    output_path = given_args.get('output_path', '')

    if output_path != '' and not output_path.exists():
        output_path.mkdir(parents=True)
    return output_path

def _check_scenario_name(given_args):
    """Ensures given output path exists.

    Arguments:
        *given_args* : dict, of keyword arguments.

    Returns:
        *str* : scenario_name
    """
    scenario_name = given_args.get('scenario_name', False)
    if not scenario_name:
        raise ValueError("Required `scenario_name` not defined.")

    return scenario_name


# Get all the needed modules
import xarray as xr
import os
import rioxarray
import numpy as np
import shapely
import pandas as pd
import geopandas as gpd
import pyproj
from tqdm import tqdm
from pathlib import PurePath
import rasterio
from exactextract import exact_extract
from pyproj import Geod
from shapely.geometry import Point, LineString

import traceback

from damagescanner.utils import _check_output_path, _check_scenario_name
from damagescanner.osm import read_osm_data


def _convert_to_meters(feature):
    """Convert coverage to meters for each individual row in the dataframe.


    Args:
        feature (gpd.GeoDataFrame): GeoDataFrame with the hazard and exposure information.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with the hazard and exposure information
        converted to meters.
    """

    line_string = feature.geometry

    # only continue if geometry is a line
    if line_string.geom_type not in ["LineString", "MultiLineString"]:
        return feature.coverage

    geod = Geod(ellps="WGS84")
    coverage_meters = []
    for cover in feature.coverage:
        new_for_length = LineString(
            [Point(line_string.coords[0]), line_string.interpolate(cover)]
        )
        coverage_meters.append(geod.geometry_length(new_for_length))

    return coverage_meters


def _create_grid(bbox, height):
    """Create a vector-based grid

    Args:
        bbox ([type]): [description]
        height ([type]): [description]

    Returns:
        [type]: [description]
    """

    # set xmin,ymin,xmax,and ymax of the grid
    xmin, ymin = shapely.total_bounds(bbox)[0], shapely.total_bounds(bbox)[1]
    xmax, ymax = shapely.total_bounds(bbox)[2], shapely.total_bounds(bbox)[3]

    # estimate total rows and columns
    rows = int(np.ceil((ymax - ymin) / height))
    cols = int(np.ceil((xmax - xmin) / height))

    # set corner points
    x_left_origin = xmin
    x_right_origin = xmin + height
    y_top_origin = ymax
    y_bottom_origin = ymax - height

    # create actual grid
    res_geoms = []
    for countcols in range(cols):
        y_top = y_top_origin
        y_bottom = y_bottom_origin
        for countrows in range(rows):
            res_geoms.append(
                (
                    (
                        (x_left_origin, y_top),
                        (x_right_origin, y_top),
                        (x_right_origin, y_bottom),
                        (x_left_origin, y_bottom),
                    )
                )
            )
            y_top = y_top - height
            y_bottom = y_bottom - height
        x_left_origin = x_left_origin + height
        x_right_origin = x_right_origin + height

    # return grid as shapely polygons
    return shapely.polygons(res_geoms)


def _remove_duplicates(df):
    """Remove duplicates from a dataframe.

    Args:
        df (pd.DataFrame): DataFrame with duplicates.

    Returns:
        pd.DataFrame: DataFrame without duplicates.
    """

    no_duplicates = []
    for row in df.groupby(level=0):
        if len(row[1]) == 1:
            no_duplicates.append(
                [
                    row[0],
                    row[1]["coverage"].values[0],
                    row[1]["values"].values[0],
                ]
            )
        elif len(row[1]) > 1:
            # concatenate coverage column lists
            import itertools

            cov_all = list(itertools.chain.from_iterable(row[1]["coverage"].values))
            # concatenate values column lists
            val_all = list(itertools.chain.from_iterable(row[1]["values"].values))

            # append to no_duplicates with new cov_sum and val_sum
            no_duplicates.append([row[0], cov_all, val_all])

    df_no_duplicates = pd.DataFrame(
        no_duplicates, columns=["index", "coverage", "values"]
    ).set_index("index")

    return df_no_duplicates


def _reproject(hazard, features, hazard_crs):
    """
    Function to reproject to the same CRS in meters.

    Args:


    Returns:
        _type_: _description_
    """

    bounds = features.total_bounds

    bbox = shapely.box(bounds[0], bounds[1], bounds[2], bounds[3])

    centre_point = bbox.centroid
    lat = shapely.get_y(centre_point)
    lon = shapely.get_x(centre_point)

    # formula below based on :https://gis.stackexchange.com/a/190209/80697

    approximate_crs = "EPSG:" + str(
        int(32700 - np.round((45 + lat) / 90, 0) * 100 + np.round((183 + lon) / 6, 0))
    )

    # reproject if needed

    if hazard_crs.to_epsg() == features.crs.to_epsg():
        return hazard, features, approximate_crs

    else:
        features = features.to_crs(hazard_crs)
        return hazard, features, approximate_crs

    # if not pyproj.CRS.from_epsg(hazard_crs.to_epsg()).axis_info[0].unit_name == "metre":
    #     hazard = hazard.rio.reproject(approximate_crs)

    # if (
    #     not pyproj.CRS.from_epsg(features.crs.to_epsg()).axis_info[0].unit_name
    #     == "metre"
    # ):
    #     features = features.to_crs(approximate_crs)

    # return hazard, features, approximate_crs


def _overlay_raster_vector(
    hazard, features, hazard_crs, hazard_col="band_data", nodata=-9999, gridded=True
):
    """
    Function to overlay a raster with a vector file.

    Args:
        hazard (xarray.DataArray): Raster file with hazard information.
        exposure (gpd.GeoDataFrame): Vector file with exposure information.
        hazard_col (str): Name of the column in the hazard file that contains the hazard intensity.
        hazard_crs (str): CRS of the hazard file.
        nodata (int, optional): Value that indicates no data in the raster file. Defaults to -9999.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with the hazard and exposure information.
    """
    # make sure the hazard data has a crs
    if hazard_crs.to_epsg() == None:
        RuntimeWarning(
            "Hazard crs is not correctly defined. We will now assume it is EPSG:4326"
        )
        hazard = hazard.rio.set_crs("EPSG:4326")
        hazard_crs = pyproj.CRS.from_epsg(4326)

    hazard[hazard_col].rio.write_nodata(nodata, inplace=True)

    area_and_line_objects = features.geom_type.isin(
        ["Polygon", "MultiPolygon", "LineString", "MultiLineString"]
    )

    point_objects = features.geom_type == "Point"

    # make sure they are both in the same coordinate system
    hazard, features, approximate_crs = _reproject(hazard, features, hazard_crs)

    # Check if the exposure data contains any area or line objects
    assert area_and_line_objects.sum() + point_objects.sum() == len(features)

    if point_objects.sum() > 0:
        # if hazard data is a rasterio object:
        if isinstance(hazard, rasterio.io.DatasetReader):
            values = np.array(
                [
                    value[0]
                    for value in rasterio.sample.sample_gen(
                        hazard,
                        [
                            (point.x, point.y)
                            for point in features[point_objects].geometry
                        ],
                    )
                ]
            )

        # if hazard data is a xarray object:
        else:
            hazard_for_points = hazard[hazard_col]
            values = hazard_for_points.sel(
                {
                    hazard_for_points.rio.x_dim: xr.DataArray(
                        features[point_objects].geometry.x
                    ),
                    hazard_for_points.rio.y_dim: xr.DataArray(
                        features[point_objects].geometry.y
                    ),
                },
                method="nearest",
            ).values[0]

        # add values to the features
        features.loc[point_objects, "values"] = values
        features.loc[point_objects, "coverage"] = [1] * len(values)

        # turn values into lists
        features.loc[point_objects, "values"] = features.loc[
            point_objects, "values"
        ].apply(lambda x: [x] if x > 0 else [])

    if not gridded:
        if area_and_line_objects.sum() > 0:
            values_and_coverage_per_area_and_line_object = exact_extract(
                hazard,
                features[area_and_line_objects],
                ["coverage", "values"],
                output="pandas",
            )

            features.loc[area_and_line_objects, "coverage"] = (
                values_and_coverage_per_area_and_line_object["coverage"].values
            )
            features.loc[area_and_line_objects, "values"] = (
                values_and_coverage_per_area_and_line_object["values"].values
            )

        # only keep features with values
        features = features[features["values"].apply(lambda x: len(x) > 0)]

        tqdm.pandas(desc="convert coverage to meters")
        features.loc[:, "coverage"] = features.progress_apply(
            lambda feature: _convert_to_meters(feature), axis=1
        )

    elif gridded:
        if area_and_line_objects.sum() > 0:
            grid_cell_size = 0.5  # in degrees

            # create grid
            bbox = shapely.box(
                hazard.rio.bounds()[0],
                hazard.rio.bounds()[1],
                hazard.rio.bounds()[2],
                hazard.rio.bounds()[3],
            )

            gridded = _create_grid(bbox, grid_cell_size)

            # get all bounds
            all_bounds = gpd.GeoDataFrame(gridded, columns=["geometry"]).bounds

            collect_overlay = []

            # loop over all grids
            for bounds in tqdm(
                all_bounds.itertuples(),
                total=len(all_bounds),
                desc="Overlay raster with vector",
            ):
                try:
                    # subset hazard
                    subset_hazard = hazard.rio.clip_box(
                        minx=bounds.minx,
                        miny=bounds.miny,
                        maxx=bounds.maxx,
                        maxy=bounds.maxy,
                    )

                    sub_bbox = shapely.box(
                        bounds.minx, bounds.miny, bounds.maxx, bounds.maxy
                    )

                    # subset features
                    subset_features = gpd.clip(features, sub_bbox)  # list(bounds)[1:])
                    subset_area_and_line_objects = subset_features.geom_type.isin(
                        ["Polygon", "MultiPolygon", "LineString", "MultiLineString"]
                    ).values

                    if len(subset_features) == 0:
                        continue

                    values_and_coverage_per_area_and_line_object = exact_extract(
                        subset_hazard,
                        subset_features[subset_area_and_line_objects],
                        ["coverage", "values"],
                        output="pandas",
                    )

                    # make sure we can connect the results with the features
                    values_and_coverage_per_area_and_line_object.index = (
                        subset_features[subset_area_and_line_objects].index
                    )
                    collect_overlay.append(values_and_coverage_per_area_and_line_object)

                except:
                    get_error = traceback.format_exc()
                    error_to_ignore = "At least one of the clipped raster x,y coordinates has only one point."
                    if error_to_ignore not in get_error:
                        traceback.print_exc()

            df = pd.concat(collect_overlay).sort_index()

            # remove duplicates
            df_no_duplicates = _remove_duplicates(df)

            ## add features to the original features
            features.loc[df_no_duplicates.index, "coverage"] = df_no_duplicates[
                "coverage"
            ]

            features.loc[df_no_duplicates.index, "values"] = df_no_duplicates["values"]

            # only keep features with values
            features = features[features["values"].apply(lambda x: len(x) > 0)]

            # convert coverage to meters, only do this if the crs is not in meters
            if (
                not pyproj.CRS.from_epsg(hazard_crs.to_epsg()).axis_info[0].unit_name
                == "metre"
            ):
                tqdm.pandas(desc="convert coverage to meters")

                features.loc[:, "coverage"] = features.progress_apply(
                    lambda feature: _convert_to_meters(feature), axis=1
                )

    return features


def _overlay_vector_vector(
    hazard, features, hazard_col="band_data", nodata=-9999, gridded=False
):
    """
    Function to overlay a vector with another vector file.

    Args:
        hazard (gpd.GeoDataFrame): Vector file with hazard information.
        exposure (gpd.GeoDataFrame): Vector file with exposure information.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with the hazard and exposure information.

    WIP: This function is not yet finished. It is currently only able to overlay point objects.
    """
    # make sure the hazard data has a crs
    if hazard_crs.to_epsg() == None:
        RuntimeWarning(
            "Hazard crs is not correctly defined. We will now assume it is EPSG:4326"
        )
        hazard = hazard.rio.set_crs("EPSG:4326")
        hazard_crs = pyproj.CRS.from_epsg(4326)

    hazard[hazard_col].rio.write_nodata(nodata, inplace=True)

    area_and_line_objects = features.geom_type.isin(
        ["Polygon", "MultiPolygon", "LineString", "MultiLineString"]
    )
    point_objects = features.geom_type == "Point"

    # Check if the exposure data contains any area or line objects
    assert area_and_line_objects.sum() + point_objects.sum() == len(features)

    if not gridded:
        # reproject if needed
        hazard, features, approximate_crs = _reproject(hazard, features, hazard_crs)

    return features


def _estimate_damage(features, curves, cell_area_m2):
    """
    Function to estimate the damage based on the hazard and exposure information.

    Args:
        exposure (gpd.GeoDataFrame): GeoDataFrame with the hazard and exposure information.
        curves (pd.DataFrame): DataFrame with the stage-damage curves of the different land-use classes.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with the hazard, exposure and damage information.
    """

    features["damage"] = features.progress_apply(
        lambda _object: _get_damage_per_object(_object, curves, cell_area_m2),
        axis=1,
    )
    return features


def _get_damage_per_object(asset, curves, cell_area_m2):
    """
    Calculate damage for a given asset based on hazard information.
    Arguments:
        *asset*: Tuple containing information about the asset. It includes:
            - Index or identifier of the asset (asset[0]).
            - Asset-specific information, including hazard points (asset[1]['hazard_point']).
        *maxdam_dict*: Maximum damage value.
    Returns:
        *tuple*: A tuple containing the asset index or identifier and the calculated damage.
    """

    if asset.geometry.geom_type in ("Polygon", "MultiPolygon"):
        coverage = asset["coverage"] * cell_area_m2
    elif asset.geometry.geom_type in ("LineString", "MultiLineString"):
        coverage = asset["coverage"]
    elif asset.geometry.geom_type in ("Point"):
        coverage = 1
    else:
        raise ValueError(f"Geometry type {asset.geometry.geom_type} not supported")

    return (
        np.sum(
            np.interp(
                asset["values"], curves.index, curves[asset["object_type"]].values
            )
            * coverage
        )
        * asset["maximum_damage"]
    )


def VectorExposure(hazard_file, feature_file, asset_type="roads"):
    """
    Function to assess the exposure of objects.

    Args:
        hazard_file (str): Path to the hazard file.
        exposure_file (str): Path to the exposure file.

    Returns:
        gpd.GeoDataFrame: GeoDataFrame with the hazard and exposure information.
    """
    # load exposure data
    if isinstance(feature_file, PurePath):
        # if exposure_file is a shapefile, geopackage or parquet file
        if feature_file.suffix in [".shp", ".gpkg", "parquet"]:
            features = gpd.read_file(feature_file)

        # if exposure_file is an osm.pbf file
        elif feature_file.suffix == ".pbf":
            features = read_osm_data(feature_file, asset_type)
            object_col = "object_type"
        else:
            raise ValueError(
                "exposure data should either be a shapefile, geopackage, parquet or osm.pbf file"
            )

    elif isinstance(feature_file, gpd.GeoDataFrame) | isinstance(
        feature_file, pd.DataFrame
    ):
        features = gpd.GeoDataFrame(feature_file.copy())

    else:
        raise ValueError(
            "exposure data should either be a shapefile, geopackage, parquet or osm.pbf file"
        )

    # load hazard data
    if isinstance(hazard_file, PurePath):
        if hazard_file.suffix in [".tif", ".tiff", ".nc"]:
            hazard = xr.open_dataset(hazard_file, engine="rasterio")
            hazard_crs = hazard.rio.crs
            cell_area_m2 = abs(hazard.rio.resolution()[0]) * abs(
                hazard.rio.resolution()[1]
            )
        elif hazard_file.suffix in [".shp", ".gpkg", ".pbf"]:
            hazard = gpd.read_file(hazard_file)
            hazard_crs = hazard.crs
            cell_area_m2 = 1
        else:
            raise ValueError(
                "hazard data should either be a geotiff, netcdf, shapefile, geopackage or parquet file"
            )
    elif isinstance(hazard, rasterio.io.DatasetReader):
        hazard_crs = hazard.crs
        cell_area_m2 = hazard.res[0] * hazard.res[1]
    elif isinstance(hazard, (xr.Dataset, xr.DataArray)):
        hazard_crs = hazard.rio.crs
        cell_area_m2 = abs(hazard.rio.resolution()[0]) * abs(hazard.rio.resolution()[1])
    elif isinstance(hazard, gpd.GeoDataFrame):
        hazard_crs = hazard.crs
        cell_area_m2 = 1
    else:
        raise ValueError(
            f"Hazard should be a raster or GeoDataFrame object, {type(hazard)} given"
        )

    # Run exposure overlay
    if isinstance(hazard, (rasterio.io.DatasetReader, xr.Dataset, xr.DataArray)):
        features = _overlay_raster_vector(hazard, features, hazard_crs)
    elif isinstance(hazard, (gpd.GeoDataFrame, pd.DataFrame)):
        features = _overlay_vector_vector(hazard, features)

    return features, object_col, hazard_crs, cell_area_m2


def VectorScanner(
    hazard_file,
    feature_file,
    curve_path,
    maxdam_path,
    asset_type="roads",
    save=False,
    **kwargs,
):
    """
    Vector based implementation of a direct damage assessment

    Arguments:
        *exposure_file* : Shapefile, Pandas DataFrame or Geopandas GeoDataFrame
        with land-use information of the area.

        *hazard_file* : GeoTiff with inundation depth per grid cell. Make sure
        that the unit of the inundation map corresponds with the unit of the
        first column of the curves file.

        *curve_path* : File with the stage-damage curves of the different
        land-use classes. Can also be a pandas DataFrame (but not a numpy Array).

        *maxdam_path* : File with the maximum damages per land-use class
        (in euro/m2). Can also be a pandas DataFrame (but not a numpy Array).

    Optional Arguments:

        *object_col* : Specify the column name of the unique object id's.
        Default is set to **landuse**.

        *sub_types* : List of subtypes that should be included in the analysis.

        *save* : Set to True if you would like to save the output. Requires
        several **kwargs**

    kwargs:
        *output_path* : Specify where files should be saved.

        *scenario_name*: Give a unique name for the files that are going to be saved.

    Raises:
        *ValueError* : on missing kwargs

    Returns:
     *damagebin* : Table with the land-use class names (1st column) and the
     damage for that land-use class (2nd column).

    TO DO: add the option to use vector data for the hazard as well.

    """

    # Load hazard and exposure data, and perform the overlay
    features, object_col, hazard_crs, cell_area_m2 = VectorExposure(
        hazard_file, feature_file, asset_type
    )

    # Load curves
    if isinstance(curve_path, pd.DataFrame):
        curves = curve_path.copy()
    elif isinstance(curve_path, np.ndarray):
        raise ValueError(
            "For the vector-based approach we use a pandas DataFrame, not a Numpy Array"
        )
    elif curve_path.parts[-1].endswith(".csv"):
        curves = pd.read_csv(curve_path, index_col=[0])

    # Load maximum damages
    if isinstance(maxdam_path, PurePath) and maxdam_path.parts[-1].endswith(".csv"):
        maxdam = pd.read_csv(maxdam_path)
        maxdam = dict(zip(maxdam["object_type"], maxdam["damage"]))
    elif isinstance(maxdam_path, pd.DataFrame):
        maxdam = dict(zip(maxdam_path["object_type"], maxdam_path["damage"]))
    elif isinstance(maxdam_path, np.ndarray):
        maxdam = dict(zip(maxdam_path[:, 0], maxdam_path[:, 1]))
    elif isinstance(maxdam_path, dict):
        maxdam = maxdam_path

    # convert exposure column with asset types to object_type
    features = features.rename(columns={object_col: "object_type"})

    # connect maxdam to exposure
    try:
        features["maximum_damage"] = features.apply(
            lambda x: maxdam[x["object_type"]], axis=1
        )
    except KeyError:
        raise KeyError(
            "Not all object types in the exposure are included in the maximum damage file"
        )

    # Calculate damage
    tqdm.pandas(desc="Calculating damage")
    features = _estimate_damage(features, curves, cell_area_m2)

    # # Save output
    # if save == True:
    #     # requires adding output_path and scenario_name to function call
    #     # If output path is not defined, will place file in current directory
    #     output_path = _check_output_path(kwargs)
    #     scenario_name = _check_scenario_name(kwargs)
    #     path_prefix = PurePath(output_path, scenario_name)

    #     damage_fn = f'{path_prefix}_damages.csv'
    #     damaged_objects.to_csv(damage_fn)
    #     return damaged_objects

    # else:
    #     return damaged_objects

    return features



"""DamageScanner - a directe damage assessment toolkit

Raster specific functions
"""

# Get all the needed modules
from osgeo import gdal
import rasterio
import xarray as xr
import numpy as np
import pandas as pd
from affine import Affine
import warnings
from pathlib import PurePath
import shapely
from damagescanner.utils import _check_output_path, _check_scenario_name


def _match_rasters(raster_in1, raster_in2):
    """
    In case of a mismatch between two rasters, return only the intersecting parts.

    Code adapted from http://sciience.tumblr.com/post/101722591382/finding-the-georeferenced-intersection-between-two

    Arguments:
        *raster_in1* : One of the two rasters to be clipped to the overlapping extent.

        *raster_in2* : One of the two rasters to be clipped to the overlapping extent.

    Returns:
        *array1* : Numpy Array of raster1

        *array2* : Numpy Array of raster2

        *intersection* : Bounding box of overlapping part

    """
    # Read rasterdata
    raster1 = gdal.Open(raster_in1)
    raster2 = gdal.Open(raster_in2)

    # load data
    band1 = raster1.GetRasterBand(1)
    band2 = raster2.GetRasterBand(1)
    gt1 = raster1.GetGeoTransform()
    gt2 = raster2.GetGeoTransform()

    # find each image's bounding box
    # r1 has left, top, right, bottom of dataset's bounds in geospatial coordinates.
    r1 = [
        gt1[0],
        gt1[3],
        gt1[0] + (gt1[1] * raster1.RasterXSize),
        gt1[3] + (gt1[5] * raster1.RasterYSize),
    ]
    r2 = [
        gt2[0],
        gt2[3],
        gt2[0] + (gt2[1] * raster2.RasterXSize),
        gt2[3] + (gt2[5] * raster2.RasterYSize),
    ]
    print("\t1 bounding box: %s" % str(r1))
    print("\t2 bounding box: %s" % str(r2))

    # find intersection between bounding boxes
    intersection = [
        max(r1[0], r2[0]),
        min(r1[1], r2[1]),
        min(r1[2], r2[2]),
        max(r1[3], r2[3]),
    ]
    if r1 != r2:
        print("\t** different bounding boxes **")
        # check for any overlap at all...
        if (intersection[2] < intersection[0]) or (intersection[1] < intersection[3]):
            intersection = None
            print("\t***no overlap***")
            return
        else:
            print("\tintersection:", intersection)
            left1 = int(
                round((intersection[0] - r1[0]) / gt1[1])
            )  # difference divided by pixel dimension
            top1 = int(round((intersection[1] - r1[1]) / gt1[5]))
            col1 = (
                int(round((intersection[2] - r1[0]) / gt1[1])) - left1
            )  # difference minus offset left
            row1 = int(round((intersection[3] - r1[1]) / gt1[5])) - top1

            left2 = int(
                round((intersection[0] - r2[0]) / gt2[1])
            )  # difference divided by pixel dimension
            top2 = int(round((intersection[1] - r2[1]) / gt2[5]))
            col2 = (
                int(round((intersection[2] - r2[0]) / gt2[1])) - left2
            )  # difference minus new left offset
            row2 = int(round((intersection[3] - r2[1]) / gt2[5])) - top2

            # print '\tcol1:',col1,'row1:',row1,'col2:',col2,'row2:',row2
            if col1 != col2 or row1 != row2:
                print("ERROR: Columns and rows still do not match! ***")
            # these arrays should now have the same spatial geometry though NaNs may differ
            array1 = band1.ReadAsArray(left1, top1, col1, row1)
            array2 = band2.ReadAsArray(left2, top2, col2, row2)

    else:  # same dimensions from the get go
        col1 = raster1.RasterXSize  # = col2
        row1 = raster1.RasterYSize  # = row2
        array1 = band1.ReadAsArray()
        array2 = band2.ReadAsArray()

    return array1, array2, intersection

def _match_raster_to_vector(hazard, landuse, lu_crs, haz_crs, resolution, hazard_col):
    """Matches the resolution and extent of a raster to a vector file.

    Arguments:
        *hazard* : netCDF4 with hazard intensity per grid cell.
        *landuse* : netCDF4 with land-use information per grid cell.
        *lu_crs* : EPSG code of the land-use file.
        *haz_crs* : EPSG code of the hazard file.
        *resolution* : Desired resolution of the raster file.
        *hazard_col* : Name of the column in the hazard file that contains the hazard intensity.

    Returns:
        *hazard* : DataSet with hazard intensity per grid cell.

        *landuse* : DataSet with land-use information per grid cell.

    """
    # Set the crs of the hazard variable to haz_crs
    hazard.rio.write_crs(haz_crs, inplace=True)

    # Rename the latitude and longitude variables to 'y' and 'x' respectively
    hazard = hazard.rename({"Latitude": "y", "Longitude": "x"})

    # Set the x and y dimensions in the hazard variable to 'x' and 'y' respectively
    hazard.rio.set_spatial_dims(x_dim="x", y_dim="y", inplace=True)

    # Set the crs of the landuse variable to lu_crs
    landuse.rio.write_crs(lu_crs, inplace=True)

    # Reproject the landuse variable from EPSG:4326 to EPSG:3857
    landuse = landuse.rio.reproject("EPSG:3857", resolution=resolution)

    # Get the minimum longitude and latitude values in the landuse variable
    min_lon = landuse.x.min().to_dict()["data"]
    min_lat = landuse.y.min().to_dict()["data"]

    # Get the maximum longitude and latitude values in the landuse variable
    max_lon = landuse.x.max().to_dict()["data"]
    max_lat = landuse.y.max().to_dict()["data"]

    # Create a bounding box using the minimum and maximum latitude and longitude values
    area = gpd.GeoDataFrame(
        [shapely.box(min_lon, min_lat, max_lon, max_lat)], columns=["geometry"]
    )

    # Set the crs of the bounding box to EPSG:3857
    area.crs = "epsg:3857"

    # Convert the crs of the bounding box to EPSG:4326
    area = area.to_crs("epsg:4326")

    # Clip the hazard variable to the extent of the bounding box
    hazard = hazard.rio.clip(area.geometry.values, area.crs)

    # Reproject the hazard variable to EPSG:3857 with the desired resolution
    hazard = hazard.rio.reproject("EPSG:3857", resolution=resolution)

    # Clip the hazard variable again to the extent of the bounding box
    hazard = hazard.rio.clip(area.geometry.values, area.crs)

    # If the hazard variable has fewer columns and rows than the landuse variable, reproject
    # the landuse variable to match the hazard variable
    if (len(hazard.x) < len(landuse.x)) & (len(hazard.y) < len(landuse.y)):
        landuse = landuse.rio.reproject_match(hazard)

    # If the hazard variable has more columns and rows than the landuse variable,
    # reproject the hazard variable to match the landuse variable

    elif (len(hazard.x) > len(landuse.x)) & (len(hazard.y) > len(landuse.y)):
        hazard = hazard.rio.reproject_match(landuse)

    # Convert the hazard and landuse variable to a numpy array
    landuse = landuse["band_data"].to_numpy()[0, :, :]
    hazard = hazard[hazard_col].to_numpy()[0, :, :]

    return hazard, landuse

def RasterScanner(
    exposure_file,
    hazard_file,
    curve_path,
    maxdam_path,
    lu_crs=28992,
    haz_crs=4326,
    hazard_col="FX",
    dtype=np.int32,
    save=False,
    **kwargs,
):
    """
    Raster-based implementation of a direct damage assessment.

    Arguments:
        *landuse_file* : GeoTiff with land-use information per grid cell. Make sure
        the land-use categories correspond with the curves and maximum damages
        (see below). Furthermore, the resolution and extend of the land-use map
        has to be exactly the same as the inundation map.

        *hazard_file* : GeoTiff or netCDF4 with hazard intensity per grid cell. Make sure
        that the unit of the hazard map corresponds with the unit of the
        first column of the curves file.

        *curve_path* : File with the stage-damage curves of the different
        land-use classes. Values should be given as ratios, i.e. between 0 and 1.
        Can also be a pandas DataFrame or numpy Array.

        *maxdam_path* : File with the maximum damages per land-use class
        (in euro/m2). Can also be a pandas DataFrame or numpy Array.

        *dtype*: Set the dtype to the requires precision. This will affect the output damage raster as well

    Optional Arguments:
        *save* : Set to True if you would like to save the output. Requires
        several **kwargs**

    kwargs:
        *nan_value* : if nan_value is provided, will mask the inundation file.
        This option can significantly fasten computations

        *cell_size* : If both the landuse and hazard map are numpy arrays,
        manually set the cell size.

        *resolution* : If landuse is a numpy array, but the hazard map
        is a netcdf, you need to specify the resolution of the landuse map.

        *output_path* : Specify where files should be saved.

        *scenario_name*: Give a unique name for the files that are going to be saved.

        *in_millions*: Set to True if all values should be set in millions.

        *crs*: Specify crs if you only read in two numpy array

        *transform*: Specify transform if you only read in numpy arrays in order to save the result raster

    Raises:
        *ValueError* : on missing kwarg options

    Returns:
     *damagebin* : Table with the land-use class numbers (1st column) and the
     damage for that land-use class (2nd column).

     *damagemap* : Map displaying the damage per grid cell of the area.

    """
    # load land-use map
    if isinstance(exposure_file, PurePath):
        with rasterio.open(exposure_file) as src:
            landuse = src.read()[0, :, :]
            transform = src.transform
            resolution = src.res[0]
            cellsize = src.res[0] * src.res[1]
    else:
        landuse = exposure_file.copy()

    landuse_in = landuse.copy()

    # Load hazard map
    if isinstance(hazard_file, PurePath):
        if hazard_file.parts[-1].endswith(".tif") | hazard_file.parts[-1].endswith(
            ".tiff"
        ):
            with rasterio.open(hazard_file) as src:
                hazard = src.read()[0, :, :]
                transform = src.transform

        elif hazard_file.parts[-1].endswith(".nc"):
            # Open the hazard netcdf file and store it in the hazard variable
            hazard = xr.open_dataset(hazard_file)

            # Open the landuse geotiff file and store it in the landuse variable
            landuse = xr.open_dataset(exposure_file, engine="rasterio")

            # Match raster to vector
            hazard, landuse = _match_raster_to_vector(
                hazard, landuse, lu_crs, haz_crs, resolution, hazard_col
            )

    elif isinstance(hazard_file, xr.Dataset):
        # Open the landuse geotiff file and store it in the landuse variable
        landuse = xr.open_dataset(exposure_file, engine="rasterio")

        # Match raster to vector
        hazard, landuse = _match_raster_to_vector(
            hazard_file, landuse, lu_crs, haz_crs, resolution, hazard_col
        )

    else:
        hazard = hazard_file.copy()

    # check if land-use and hazard map have the same shape.
    if landuse.shape != hazard.shape:
        warnings.warn(
            "WARNING: landuse and hazard maps are not the same shape. Let's fix this first!"
        )

        landuse, hazard, intersection = _match_rasters(exposure_file, hazard_file)

        # create the right affine for saving the output
        transform = Affine(
            transform[0],
            transform[1],
            intersection[0],
            transform[3],
            transform[4],
            intersection[1],
        )

    # set cellsize:
    if isinstance(exposure_file, PurePath) | isinstance(hazard_file, PurePath):
        cellsize = src.res[0] * src.res[1]
    else:
        try:
            cellsize = kwargs["cellsize"]
        except KeyError:
            raise ValueError("Required `cellsize` not given.")

    # Load curves
    if isinstance(curve_path, pd.DataFrame):
        curves = curve_path.values
    elif isinstance(curve_path, np.ndarray):
        curves = curve_path
    elif curve_path.parts[-1].endswith(".csv"):
        curves = pd.read_csv(curve_path).values

    if ((curves > 1).all()) or ((curves < 0).all()):
        raise ValueError("Stage-damage curve values must be between 0 and 1")

    # Load maximum damages
    if isinstance(maxdam_path, pd.DataFrame):
        maxdam = maxdam_path.values
    elif isinstance(maxdam_path, np.ndarray):
        maxdam = maxdam_path
    elif maxdam_path.parts[-1].endswith(".csv"):
        maxdam = pd.read_csv(maxdam_path).values

    if maxdam.shape[0] != (curves.shape[1] - 1):
        raise ValueError(
            "Dimensions between maximum damages and the number of depth-damage curve do not agree"
        )

    # Speed up calculation by only considering feasible points
    if kwargs.get("nan_value"):
        nan_value = kwargs.get("nan_value")
        hazard[hazard == nan_value] = 0

    haz = hazard * (hazard >= 0) + 0
    haz[haz >= curves[:, 0].max()] = curves[:, 0].max()
    area = haz > 0
    haz_intensity = haz[haz > 0]
    landuse = landuse[haz > 0]

    # Calculate damage per land-use class for structures
    numberofclasses = len(maxdam)
    alldamage = np.zeros(landuse.shape[0])
    damagebin = np.zeros(
        (
            numberofclasses,
            2,
        )
    )
    for i in range(0, numberofclasses):
        n = maxdam[i, 0]
        damagebin[i, 0] = n
        wd = haz_intensity[landuse == n]
        alpha = np.interp(wd, (curves[:, 0]), curves[:, i + 1])
        damage = alpha * (maxdam[i, 1] * cellsize)
        damagebin[i, 1] = sum(damage)
        alldamage[landuse == n] = damage

    # create the damagemap
    damagemap = np.zeros((area.shape[0], area.shape[1]), dtype=dtype)
    damagemap[area] = alldamage

    # create pandas dataframe with output
    damage_df = (
        pd.DataFrame(damagebin.astype(dtype), columns=["landuse", "damages"])
        .groupby("landuse")
        .sum()
    )

    if save:
        crs = kwargs.get("crs", src.crs)
        transform = kwargs.get("transform", transform)

        # requires adding output_path and scenario_name to function call
        # If output path is not defined, will place file in current directory
        output_path = _check_output_path(kwargs)
        scenario_name = _check_scenario_name(kwargs)
        path_prefix = PurePath(output_path, scenario_name)

        damage_fn = "{}_damages.csv".format(path_prefix)
        damage_df.to_csv(damage_fn)

        dmap_fn = "{}_damagemap.tif".format(path_prefix)
        rst_opts = {
            "driver": "GTiff",
            "height": damagemap.shape[0],
            "width": damagemap.shape[1],
            "count": 1,
            "dtype": dtype,
            "crs": crs,
            "transform": transform,
            "compress": "LZW",
        }
        with rasterio.open(dmap_fn, "w", **rst_opts) as dst:
            dst.write(damagemap, 1)

    if "in_millions" in kwargs:
        damage_df = damage_df / 1e6

    # return output
    return damage_df, damagemap, landuse_in, hazard



"""Test core objects/concepts
"""
import unittest
import numpy as np
import pandas as pd
from damagescanner.core import RasterScanner  
from pathlib import Path

class TestRasterScanner(unittest.TestCase):

    def test_raster_scanner(self):
        # Define paths to example files
        data_path = Path(__file__).parent.parent / 'data'
        landuse_file = data_path / 'landuse' / 'landuse_map.tif'
        hazard_file = data_path / 'hazard' / 'inundation_map.tif'
        curve_path = data_path / 'curves' / 'curves.csv'
        maxdam_path = data_path / 'curves' / 'maxdam.csv'

        # Call the RasterScanner function
        damage_df, damagemap, landuse_in, hazard = RasterScanner(
            landuse_file=landuse_file,
            hazard_file=hazard_file,
            curve_path=curve_path,
            maxdam_path=maxdam_path,
            lu_crs=28992,
            haz_crs=4326,
            dtype=np.int32,
            save=False
            # Add any other required parameters or kwargs here
        )

        # Perform assertions to check the correctness of the results
        self.assertIsInstance(damage_df, pd.DataFrame)
        self.assertIsInstance(damagemap, np.ndarray)
        self.assertIsInstance(landuse_in, np.ndarray)
        self.assertIsInstance(hazard, np.ndarray)

        # Add more specific assertions based on your requirements

if __name__ == "__main__":
    TESTS = unittest.TestLoader().loadTestsFromTestCase(TestRasterScanner)
    unittest.TextTestRunner(verbosity=2).run(TESTS)


"""DamageScanner - a directe damage assessment toolkit

Copyright (C) 2023 Elco Koks. All versions released under the MIT license.
"""

# Get all the needed modules
import shapely
import xarray as xr
import pandas as pd
import geopandas as gpd
from tqdm import tqdm
from pathlib import Path
from scipy import integrate

from damagescanner.vector import VectorScanner, VectorExposure
from damagescanner.raster import RasterScanner

import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)


class DamageScanner(object):
    """DamageScanner - a directe damage assessment toolkit"""

    def __init__(self, hazard_data, feature_data, curves, maxdam):
        """Prepare the input for a damage assessment"""

        # Collect the input data
        self.hazard_data = hazard_data
        self.feature_data = feature_data
        self.curves = curves
        self.maxdam = maxdam

        # Convert the input to a Path object if it is a string
        if isinstance(self.feature_data, str):
            self.feature_data = Path(feature_data)

        if isinstance(self.hazard_data, str):
            self.hazard_data = Path(hazard_data)

        if isinstance(self.curves, str):
            self.curves = Path(curves)

        if isinstance(self.maxdam, str):
            self.maxdam = Path(maxdam)

        # Check the type of the exposure data
        if isinstance(self.feature_data, Path):
            if self.feature_data.suffix in [".tif", ".tiff", ".nc"]:
                self.feature_data = "raster"

            elif self.feature_data.suffix in [
                ".shp",
                ".gpkg",
                ".pbf",
                ".geofeather",
                ".geoparquet",
            ]:
                self.assessment_type = "vector"

                if self.feature_data.suffix == ".pbf":
                    self.osm = True
            else:
                raise ImportError(
                    ": The exposure data should be a a shapefile, geopackage, \
                        geoparquet, osm.pbf, geotiff or netcdf file."
                )

        else:
            if isinstance(self.feature_data, (xr.DataArray, xr.Dataset)):
                self.assessment_type = "raster"
            elif isinstance(self.feature_data, (gpd.GeoDataFrame, pd.DataFrame)):
                self.assessment_type = "vector"

        # Collect vulnerability curves
        if isinstance(curves, (pd.DataFrame, Path)):
            self.curves = curves
        else:
            raise ImportWarning(
                "Prepare the vulnerability curves as a pandas DataFrame or a \
                     as a directory path to a csv file"
            )

        # Collect maxdam information
        if isinstance(maxdam, (pd.DataFrame, Path)):
            self.maxdam = maxdam
        else:
            raise ImportWarning(
                "Prepare the maximum damages as a pandas DataFrame or or a\
                     as a directory path to a csv file"
            )

    def exposure(self, **kwargs):
        """Exposure data"""

        if self.assessment_type == "raster":
            return xr.open_rasterio(self.exposure_data)

        elif self.assessment_type == "vector":
            # specificy essential data input characteristics
            if "asset_type" in kwargs:
                self.asset_type = kwargs.get("asset_type")
            else:
                self.asset_type = "landuse"

            exposed_assets = VectorExposure(
                hazard_file=self.hazard_data,
                feature_file=self.feature_data,
                asset_type=self.asset_type,
            )[0]

            return exposed_assets

    def calculate(self, save_output=False, **kwargs):
        """Damage assessment. Can be a specific hazard event, or a specific \
            single hazard footprint, or a list of events/footprints."""

        if not hasattr(self, "assessment_type"):
            raise ImportError("Please prepare the input data first")

        if self.assessment_type == "raster":
            return RasterScanner(
                exposure_file=self.feature_data,
                hazard_file=self.hazard_data,
                curve_path=self.curves,
                maxdam_path=self.maxdam,
                save=save_output,
            )

        elif self.assessment_type == "vector":
            # specificy essential data input characteristics
            if "asset_type" in kwargs:
                self.asset_type = kwargs.get("asset_type")
            else:
                self.asset_type = "landuse"

            return VectorScanner(
                hazard_file=self.hazard_data,
                feature_file=self.feature_data,
                curve_path=self.curves,
                maxdam_path=self.maxdam,
                asset_type=self.asset_type,  #'landuse',
                sub_types=kwargs.get("subtypes", None),
                save=save_output,
            )

    def risk(hazard_dict, exposure, curves, maxdam):
        """
        Calculate the risk for a list of hazard events
        """

        RP_list = list(hazard_dict.keys())

        risk = {}
        for key, value in tqdm(
            hazard_dict.items(), total=len(hazard_dict), desc="Risk Calculation"
        ):
            risk[key] = DamageScanner(value, exposure, curves, maxdam).calculate()[0]

        df_risk = pd.concat(risk, axis=1)

        RPS = [1 / x for x in RP_list]

        risk = pd.DataFrame(
            df_risk.apply(
                lambda x: integrate.simpson(y=x[RP_list][::-1], x=RPS[::-1]), axis=1
            ),
            columns=["tot_risk"],
        )

        return risk


if __name__ == "__main__":
    ####################################################################################################

    # Kampen

    data_path = Path("..") / ".." / "data" / "kampen"

    # define the input data
    exposure = data_path / "exposure" / "landuse_map.tif"
    hazard = data_path / "hazard" / "1in100_inundation_map.tif"
    curves = data_path / "vulnerability" / "curves.csv"
    maxdam = data_path / "vulnerability" / "maxdam.csv"

    # initiate the damage scanner and calculate the damages
    # print(DamageScanner(hazard, exposure, curves, maxdam).calculate()[0])

    # define the input data
    exposure = data_path / "exposure" / "landuse.shp"
    hazard = data_path / "hazard" / "1in100_inundation_map.tif"
    curves = data_path / "vulnerability" / "curves_osm.csv"
    maxdam = data_path / "vulnerability" / "maxdam_osm.csv"

    # initiate the damage scanner and calculate the damages
    # print(DamageScanner(hazard, exposure, curves, maxdam).calculate())

    ####################################################################################################
    # Kampen risk assessment
    data_path = Path("..") / ".." / "data" / "kampen"

    # define the input data
    hazard_dict = {
        10: data_path / "hazard" / "1in10_inundation_map.tif",
        50: data_path / "hazard" / "1in50_inundation_map.tif",
        100: data_path / "hazard" / "1in100_inundation_map.tif",
        500: data_path / "hazard" / "1in500_inundation_map.tif",
        1000: data_path / "hazard" / "1in1000_inundation_map.tif",
    }

    exposure = data_path / "exposure" / "landuse_map.tif"
    curves = data_path / "vulnerability" / "curves.csv"
    maxdam = data_path / "vulnerability" / "maxdam.csv"

    # calculate the risk
    # print(DamageScanner.risk(hazard_dict, exposure, curves, maxdam))

    ####################################################################################################
    # Jamaica

    data_path = Path("..") / ".." / "data" / "jamaica"

    # define the input data
    features = data_path / "exposure" / "jamaica-latest.osm.pbf"
    hazard = data_path / "hazard" / "FD_1in1000.tif"
    curves = data_path / "vulnerability" / "curves_osm.csv"
    maxdam = data_path / "vulnerability" / "maxdam_osm.csv"

    # estimate exposure
    asset_types = [
        "main_roads",
        "rail",
        "air",
        "telecom",
        "water_supply",
        "waste_solid",
        "waste_water",
        "education",
        "healthcare",
        "power",
        "gas",
        "food",
        "oil",
        "wastewater",
        "buildings",
    ]

    for asset_type in asset_types:
        exposed_features = DamageScanner(hazard, features, curves, maxdam).exposure(
            asset_type=asset_type
        )

        # exposed_features.to_parquet("main_roads.parquet")
        print(exposed_features[["object_type", "coverage", "values"]])

    # #initiate the damage scanner and calculate the damages
    print(
        DamageScanner(hazard, features, curves, maxdam)
        .calculate(asset_type="main_roads")
        .damage.sum()
    )
